class Section:
    def __init__(self, name, **kwargs):
        self.name = name
        self.parameters = kwargs

    def get_parameters(self):
        return [(f"{self.name}:{k}", v) for k, v in self.parameters.items()]



#!/usr/bin/env python3

### Warnings ###
import warnings

warnings.filterwarnings("ignore")

### System ###
import os
import sys
import json
import pickle
import shutil
import inspect
import configparser
from glob import glob
from hashlib import md5
from collections import defaultdict

### CLI ###
import click

### Display ###
from termcolor import colored

### Logging ###
import logging
import logzero
from logzero import logger, LogFormatter

### Data Handling ###
import numpy as np

# import pandas as pd

### Hyperparameter Optimization ###
from mango import scheduler, Tuner

### MoonLine ###
from linechart import Tearsheet, DailyPerformance, AggregateDailyPerformance
from moonline import (
    main as moonline_main,
    get_config as moonline_get_config,
    get_strategies_dir as moonline_get_strategies_dir,
)


space_x_config = {}
moonline_config = {}
moonline_config_custom = {}


def make_hash_md5(o):
    return md5(repr(make_hashable(o)).encode()).hexdigest()


def make_hashable(o):
    if isinstance(o, (tuple, list)):
        return tuple((make_hashable(e) for e in o))

    if isinstance(o, dict):
        return tuple(sorted((k, make_hashable(v)) for k, v in o.items()))

    if isinstance(o, (set, frozenset)):
        return tuple(sorted(make_hashable(e) for e in o))

    return o


def get_config(config_file, args=[]):
    config = configparser.ConfigParser()
    config.read(config_file)

    valid_categories = set(["MoonLine", "SpaceX", "Output", "Parameters"])

    config_object = defaultdict(dict)
    config_object_custom = defaultdict(dict)
    for section in config.sections():
        if section in valid_categories:
            for k, v in config[section].items():
                config_object[section][k] = v
        else:
            for k, v in config[section].items():
                config_object_custom[section.lower()][k] = v

    for i, item in enumerate(args):
        if i % 2 == 0:
            if ":" in item:
                section, value = item.lstrip("-").split(":")
            else:
                section, value = None, None
        elif section and value:
            config_object_custom[section.lower()][value] = item

    errors = []

    iterations = config_object["SpaceX"].get("iterations", None)
    if not iterations:
        config_object["SpaceX"]["iterations"] = 20
    else:
        try:
            config_object["SpaceX"]["iterations"] = int(iterations)
        except Exception as e:
            errors.append(("[SpaceX]->iterations", str(e), config_object["SpaceX"]["iterations"]))

    config_object["MoonLine"]["path"] = os.path.abspath(config_object["MoonLine"]["path"])
    if not os.path.isfile(config_object["MoonLine"]["path"]):
        errors.append(("[MoonLine]->path", "Not a file", config_object["MoonLine"]["path"]))

    config_object["Output"]["path"] = os.path.abspath(config_object["Output"]["path"])
    if os.path.isfile(config_object["Output"]["path"]):
        errors.append(("[Output]->path", "Output path is a file", config_object["Output"]["path"]))
    else:
        if os.path.isdir(config_object["Output"]["path"]):
            shutil.rmtree(config_object["Output"]["path"])
        os.makedirs(config_object["Output"]["path"])

    best_tearsheet = config_object["Output"].get("best_tearsheet", None)
    if best_tearsheet:
        config_object["Output"]["best_tearsheet"] = os.path.abspath(config_object["Output"]["best_tearsheet"])
    else:
        config_object["Output"]["best_tearsheet"] = None

    results = config_object["Output"].get("results", None)
    if results:
        config_object["Output"]["results"] = os.path.abspath(config_object["Output"]["results"])
    else:
        config_object["Output"]["results"] = None

    config_object["Parameters"]["path"] = os.path.abspath(config_object["Parameters"]["path"])
    if not os.path.isdir(config_object["Parameters"]["path"]):
        errors.append(("[Parameters]->path", "Not a directory", config_object["Parameters"]["path"]))
    else:
        found_parameter_files = glob(
            os.path.join(config_object["Parameters"]["path"], "**", "*.py"),
            recursive=True,
        )
        if not found_parameter_files:
            errors.append(("[Parameters]->name", "No parameter files found", None))
        if not config_object["Parameters"].get("name", None) and len(found_parameter_files) > 1:
            errors.append(
                (
                    "[Parameters]->name",
                    "More than one parameters file found but no name was specified",
                    None,
                )
            )

    if errors:
        for config_section, message, faulty_value in errors:
            if faulty_value:
                logger.error(
                    "{} | {} ({})".format(
                        colored(config_section, "magenta"),
                        colored(message, "red"),
                        faulty_value,
                    )
                )
            else:
                logger.error("{} | {}".format(colored(config_section, "magenta"), colored(message, "red")))
        sys.exit(1)

    return dict(config_object), dict(config_object_custom)


def get_parameters(parameter_library_path, parameter_name):
    parameter_files = glob(os.path.join(parameter_library_path, "**", "*.py"), recursive=True)
    tmp_dir = os.path.join(os.path.dirname(__file__), "parameters_tmp")
    if os.path.isdir(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir, exist_ok=True)
    for file in parameter_files:
        shutil.copyfile(file, os.path.join(tmp_dir, os.path.basename(file)))
    parameter_files = [
        os.path.join("parameters_tmp", os.path.basename(file))
        for file in glob(os.path.join(tmp_dir, "**", "*.py"), recursive=True)
    ]

    for file in parameter_files:
        if parameter_name == os.path.basename(file):
            try:
                module_name = ".".join(os.path.splitext(file)[0].split(os.sep))
                __import__(module_name, globals=globals(), locals=locals())
            except:
                module_name = "space_x." + ".".join(os.path.splitext(file)[0].split(os.sep))
                __import__(module_name, globals=globals(), locals=locals())
            classes = [(name, obj) for name, obj in inspect.getmembers(sys.modules[module_name], inspect.isclass)]
            for name, strategy_class in classes:
                if "Params" in name:
                    return strategy_class

    available_parameters = {}
    for file in parameter_files:
        try:
            module_name = ".".join(os.path.splitext(file)[0].split(os.sep))
            __import__(module_name, globals=globals(), locals=locals())
        except:
            module_name = "space_x." + ".".join(os.path.splitext(file)[0].split(os.sep))
            __import__(module_name, globals=globals(), locals=locals())
        classes = [(name, obj) for name, obj in inspect.getmembers(sys.modules[module_name], inspect.isclass)]
        for name, strategy_class in classes:
            if "Params" in name:
                available_parameters[name] = strategy_class

    if len(available_parameters) > 1:
        if parameter_name:
            target_class = available_parameters[parameter_name]
            return target_class
        else:
            names = ", ".join([colored(name, "magenta") for name in available_parameters.keys()])
            raise Exception(
                "More than one strategy candidate found: {}\n\t   Please select one with the 'name' option in config.ini".format(
                    names
                )
            )  # noqa: E501

    if not available_parameters:
        raise Exception("No strategies were found")

    target_class = available_parameters[list(available_parameters.keys())[0]]

    return target_class


# @scheduler.serial
@scheduler.parallel(n_jobs=4)
def moonline_objective(**kwargs):
    global space_x_config, moonline_config, moonline_config_custom
    config = defaultdict(dict)
    config_custom = defaultdict(dict)

    config.update(moonline_config)
    config_custom.update(moonline_config_custom)

    for k, v in kwargs.items():
        section, name = k.split(":")
        if section in config:
            config[section.lower()][name] = v
        config_custom[section.lower()][name] = v

    path_hash = make_hash_md5(kwargs)

    weight_path = os.path.join(space_x_config["Output"]["path"], path_hash, "weights.csv")
    tearsheet_path = os.path.join(space_x_config["Output"]["path"], path_hash, "tearsheet.csv")
    parameter_path = os.path.join(space_x_config["Output"]["path"], path_hash, "parameters.json")

    os.makedirs(os.path.dirname(weight_path), exist_ok=True)

    config["Output"]["weights"] = weight_path
    config["Output"]["tearsheet_csv"] = tearsheet_path

    logzero.loglevel(logging.CRITICAL)

    try:
        moonline_main(config, config_custom)
    except Exception as e:
        logger.error(f"Ran into an exception during the backtest, returning 0: {e}")
        return 0

    logzero.loglevel(logging.INFO)

    with open(parameter_path, "w") as f:
        json.dump(kwargs, f)

    # weights = pd.read_csv(weight_path)
    # tearsheet = pd.read_csv(tearsheet_path)

    perf = DailyPerformance.from_moonshot_csv(tearsheet_path)
    perf_agg = AggregateDailyPerformance(perf)

    score = (perf_agg.cagr - (perf_agg.cagr * perf_agg.max_drawdown)) + perf_agg.sharpe

    if np.isnan(score) or str(score) == "nan":
        score = 0

    logger.info(
        f"Score: {score:.2f} | Parameters: {kwargs} | CAGR: {perf_agg.cagr * 100:.2f}% | Sharpe: {perf_agg.sharpe} | Drawdown: {perf_agg.max_drawdown * 100:.2f}%"
    )

    return score


def early_stopping(results):
    global params
    logger.info(f"Best Parameters: {results['best_params']}")
    logger.info(f"Best Objective: {results['best_objective']}")

    if hasattr(params, "early_stopping"):
        return params.early_stopping(results)

    return False


@click.command(context_settings={"ignore_unknown_options": True, "allow_extra_args": True})
@click.option(
    "-c",
    "--config",
    "config_file",
    required=True,
    type=click.Path(dir_okay=False, exists=True, resolve_path=True),
    help="A file containing Space-X configuration options",
)
@click.option("-v", "--verbose", is_flag=True)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def cli(config_file, verbose, args):
    global space_x_config, moonline_config, moonline_config_custom, params

    formatter = LogFormatter(
        fmt="%(color)s[%(levelname)1.1s %(asctime)s]%(end_color)s %(message)s",
        datefmt="%H:%M:%S",
    )
    logzero.formatter(formatter)

    logzero.loglevel(logging.DEBUG if verbose else logging.INFO)

    space_x_config, _ = get_config(config_file)

    params = get_parameters(space_x_config["Parameters"]["path"], space_x_config["Parameters"]["name"])

    logger.info("Loading Parameters")
    params = params()
    available_params = params.get_sections()

    param_space = {}
    for section in available_params:
        for param, value in section.get_parameters():
            param_space[param] = value

    moonline_config, moonline_config_custom = moonline_get_config(space_x_config["MoonLine"]["path"], args)

    # Disable PDF tearsheet generation
    moonline_config["Output"]["tearsheet"] = None
    # Disable order generation
    moonline_config["Output"]["orders"] = None
    # Fill empty system section
    if "System" not in moonline_config:
        moonline_config["System"] = {}
    # Disable strategy copying and folder clearing (causes conflicts with parallel processing)
    moonline_config["System"]["read_only"] = True
    # Enable quiet mode to reduce log spam
    moonline_config["System"]["quiet"] = True

    # Copy strategies
    strategy_files = glob(os.path.join(moonline_config["Strategy"]["path"], "**", "*.py"), recursive=True)
    tmp_dir = moonline_get_strategies_dir()
    if os.path.isdir(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir, exist_ok=True)
    for file in strategy_files:
        shutil.copyfile(file, os.path.join(tmp_dir, os.path.basename(file)))

    logger.info(f"Running Hyperparameter Optimization with {space_x_config['SpaceX']['iterations']} iterations")
    tuner = Tuner(
        param_space,
        moonline_objective,
        {
            "initial_random": 4,
            "num_iteration": space_x_config["SpaceX"]["iterations"],
            "early_stopping": early_stopping,
        },
    )

    results = tuner.maximize()

    logger.info(f"Best Parameters: {results['best_params']}")
    logger.info(f"Best Objective: {results['best_objective']}")

    # Generate tearsheet for best run
    if space_x_config["Output"]["best_tearsheet"]:
        best_hash = make_hash_md5(results["best_params"])

        tearsheet_path = os.path.join(space_x_config["Output"]["path"], best_hash, "tearsheet.csv")

        file_no_ext = os.path.splitext(space_x_config["Output"]["best_tearsheet"])[0]
        shutil.copyfile(tearsheet_path, f"{file_no_ext}.csv")
        Tearsheet.from_moonshot_csv(f"{file_no_ext}.csv", pdf_filename=f"{file_no_ext}.pdf")

    for params_tried, score in sorted(zip(results["params_tried"], results["objective_values"]), key=lambda x: x[1]):
        logger.info(f"{params_tried}: {score}")

    if space_x_config["Output"]["results"]:
        with open(space_x_config["Output"]["results"], "wb") as f:
            pickle.dump(results, f)


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['space_x', 'space_x.utils']

package_data = \
{'': ['*']}

install_requires = \
['arm-mango>=1.1.0,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'moonline>=0.9.4,<0.10.0',
 'numpy>=1.20.2,<2.0.0']

entry_points = \
{'console_scripts': ['space-x = space_x.space_x:cli']}

setup_kwargs = {
    'name': 'space-x',
    'version': '0.3.1',
    'description': 'A hyperparameter optimizer for MoonLine.',
    'long_description': '# Space-X\nA hyperparameter optimizer for [Moonline](https://gitlab.genzai.dev/boosting-alpha/moonline).\n\n### Running\nTo run Space-X using `poetry`, use the following command:\n```bash\n$ poetry run python space-x -c <config_file>\n```\n\n### Generating Documentation\n```bash\n$ poetry run pdoc --config show_type_annotations=True --html --force space_x/\n```\n\n### Working on Documentation\n`pdoc` can be switched into hot-reload mode by appending the following:\n```bash\n$ poetry run pdoc --config show_type_annotations=True --html --force space_x/ --http :\n```\n\n## Creating A Configuration File\nIn keeping with the way MoonLine handles options, they are specified via a configuration file in [INI](https://en.wikipedia.org/wiki/INI_file) format. This has the added benefit of being able to quickly switch between various configurations simply by pointing it at a different configuration file.\n\n### Example Configuration\n```ini\n[MoonLine]\npath = moonline-config.ini\n\n[SpaceX]\niterations = 10\n\n[Output]\npath = out/\nbest_tearsheet = tearsheet_best.pdf\n\n[Parameters]\npath = params/\nname = ParamsA\n```\n\n### Sections\n#### MoonLine\nDescribes the MoonLine configuration to run.\n##### `path`\nThe path to the MoonLine file configuration file().\n\n#### Space-X\nSpace-X execution options.\n##### `iterations`\nThe number of iterations to run. This giverns how many parameter combinations can be checked. More iterations means more space for testing combinations. This is multiplied by the number of jobs if parallel execution is enabled.\nWhile the iterations will stay fixed (20 by default), if i.e. 4 worker processes are active, 4 times as many parameters will be explored within the same iteration.\n\n#### Output\nPaths for output artifacts.\n##### `path`\nThe path to a directory. If the directory exists, it will be overwritten. This will output sub-directories with MD5 hashes for names containing the output of a specific MoonLine run.\n##### `best_tearsheet`\nThe path to a file (`PDF`). If the file exists, it will be overwritten. This will output a tearsheet in PDF format containing multiple analytical charts to assess the performance of a given strategy. This tearsheet will only be generated for the run with the best parameters.\n\n#### Parameters\nDescribes the parameter space to run.\n##### `path`\nThe path to the directory containing the parameter file(s).\n##### `name`\nEither the exact filename of a parameter file inside the parameter directory (i.e. `crypto_parameters.py`) or the class name of a parameter definition.\n\nIf given a class name, Space-X will automatically look for and instantiate the given parameter class, so you can freely refactor your parameters and rename their files and it will still be able to pick it up.\n\nIf given a filename, only a single parameter class can be contained in the given file. When a class name is given, multiple parameters can be housed in the same file.\n\n## Usage\n```bash\nUsage: space-x [OPTIONS] [ARGS]...\n\nOptions:\n  -c, --config FILE  A file containing Space-X configuration options\n                     [required]\n\n  -v, --verbose\n  --help             Show this message and exit.\n```\n',
    'author': 'Tim Wedde',
    'author_email': 'timwedde@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.genzai.dev/boosting-alpha/space-x',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
